import json
import autograd as ag
import autograd.numpy as np
import scipy, scipy.stats
import logging
import collections as co
import pandas as pd
import matplotlib as mpl
from matplotlib import pyplot as plt
import seaborn
from .sfs_stats import JackknifeGoodnessFitStat
from .data.configurations import build_config_list
from .data.sfs import Sfs
from .demography import Demography
from .likelihood import SfsLikelihoodSurface
from .compute_sfs import expected_total_branch_len, expected_sfs, expected_heterozygosity
from .confidence_region import _ConfidenceRegion
from .events import LeafEvent, SizeEvent, JoinEvent, PulseEvent, GrowthEvent
from .events import Parameter, ParamsDict
from .events import _build_demo_graph
from .sfs_stats import SfsModelFitStats


class DemographicModel(object):
    """Object for representing and inferring a demographic history.

    :param float N_e: the population size, unless manually changed by \
    :meth:`DemographicModel.set_size()`

    :param float gen_time: units of time per generation. \
    For example, if you wish to specify time in years, and a \
    generation is 29 years, set this to 29. Default value is 1.

    :param float,None muts_per_gen: mutation rate per base \
    per generation. If unknown, set to None (the default). \
    Can be changed with :meth:`DemographicModel.set_mut_rate`
    """
    def __init__(self, N_e, gen_time=1, muts_per_gen=None):
        self.N_e = N_e
        self.gen_time = gen_time
        self.muts_per_gen = muts_per_gen

        self.parameters = co.OrderedDict()
        self.topology_events = []
        self.size_events = []
        self.leaf_events = []
        self.leafs = []

        self._set_data(sfs=None, length=None,
                       mem_chunk_size=None,
                       use_pairwise_diffs=None,
                       non_ascertained_pops=None)

    def set_mut_rate(self, muts_per_gen):
        """Set the mutation rate.

        :param float,None muts_per_gen: Mutation rate per base per generation. If unknown, set to None.
        """
        self.muts_per_gen = muts_per_gen

    def copy(self):
        ret = DemographicModel(self.N_e, self.gen_time,
                               self.muts_per_gen)
        for k, v in self.parameters.items():
            ret.parameters[k] = v.copy()
        ret.topology_events.extend(self.topology_events)
        ret.size_events.extend(self.size_events)
        ret.leaf_events.extend(self.leaf_events)
        ret.leafs.extend(self.leafs)
        ret._set_data(sfs=self._fullsfs, length=self._length,
                      mem_chunk_size=self._mem_chunk_size,
                      use_pairwise_diffs=self._use_pairwise_diffs,
                      non_ascertained_pops=self._non_ascertained_pops)
        return ret

    def set_params(self, new_params=None, randomize=False,
                   scaled=False):
        """Set the current parameter values

        :param dict/list new_params: dict mapping parameter names to \
        new values, or list of new values whose length is the current \
        number of parameters

        :param bool randomize: if True, parameters not in \
        ``new_params`` get randomly sampled new values

        :param bool scaled: if True, values in ``new_params`` have \
        been pre-scaled according to the internal representation used \
        by momi during optimization \
        (see also: :meth:`DemographicModel.add_parameter`)
        """
        if new_params is None:
            new_params = {}

        try:
            new_params.items()
        except AttributeError:
            if len(new_params) != len(self.parameters):
                raise ValueError("New parameters should either be a dict, or be a list with length current number of parameters")
            new_params = dict(zip(self.parameters.keys(), new_params))

        new_params = dict(new_params)
        curr_params = ParamsDict()
        for name, param in self.parameters.items():
            try:
                val = new_params.pop(name)
            except KeyError:
                if randomize:
                    param.resample(curr_params)
            else:
                if scaled:
                    param.x = val
                else:
                    param.x = param.inv_transform_x(val, curr_params)

            param.update_params_dict(curr_params)

        if new_params:
            raise ValueError("Unrecognized parameters: {}".format(
                list(new_params.keys())))

    def add_parameter(self, name, start_value=None,
                      rgen=None,
                      scale_transform=None,
                      unscale_transform=None,
                      scaled_lower=None,
                      scaled_upper=None):
        """Low-level method to add a new parameter. Most users should instead call
        :meth:`DemographicModel.add_size_param`,
        :meth:`DemographicModel.add_pulse_param`,
        :meth:`DemographicModel.add_time_param`, or
        :meth:`DemographicModel.add_growth_param`.

        In order for internal gradients to be correct, ``scale_transform`` and ``unscale_transform`` should be constructed using `autograd <https://github.com/HIPS/autograd/blob/master/docs/tutorial.md>`_.

        :param str name: Name of the parameter.
        :param float start_value: Starting value. If None, use \
        ``rgen`` to sample a random starting value.

        :param function rgen: Function to get a random starting value.
        :param function scale_transform: Function for internally \
        transforming and rescaling the parameter during optimization.

        :param function unscale_transform: Inverse function of \
        ``scale_transform``
        :param float scaled_lower: Lower bound after scaling by \
        ``scale_transform``
        :param float scaled_upper: Upper bound after scaling by \
        ``scale_transform``
        """
        self._conf_region = None

        assert (scale_transform is None) == (unscale_transform is None)
        if scale_transform is None:
            unscale_transform = scale_transform = lambda x, p: x

        curr_params = self.get_params()
        x_lower = scaled_lower
        x_upper = scaled_upper

        if rgen is None and start_value is None:
            raise ValueError("At least one of rgen, start_value must be specified")
        elif rgen is None:
            def rgen(params):
                return start_value
        elif start_value is None:
            start_value = rgen(curr_params)

        x0 = scale_transform(start_value, curr_params)

        param = Parameter(
            name, x0,
            transform_x=unscale_transform,
            inv_transform_x=scale_transform,
            x_bounds=(x_lower, x_upper),
            rgen=rgen)
        assert name not in self.parameters
        self.parameters[name] = param

    def add_size_param(self, name, N0=None, lower=1, upper=1e10, rgen=None):
        """Add a size parameter to the demographic model.

        :param str name: Parameter name
        :param float N0: Starting value. If None, use ``rgen`` to \
        randomly sample
        :param float lower: Lower bound
        :param float upper: Upper bound
        :param function rgen: Function to sample a random starting \
        value. If None, a truncated exponential with rate ``1 / N_e``
        """
        if rgen is None:
            scale = self.N_e
            truncexpon = scipy.stats.truncexpon(b=(upper-lower)/scale,
                                                loc=lower, scale=scale)
            def rgen(params):
                return truncexpon.rvs()

        def scale_transform(x, p):
            return np.log(x)

        self.add_parameter(name, N0,
                           scaled_lower=scale_transform(lower, None),
                           scaled_upper=scale_transform(upper, None),
                           scale_transform=scale_transform,
                           unscale_transform=lambda x, p: np.exp(x),
                           rgen=rgen)

    def add_time_param(self, name, t0=None,
                       lower=0.0, upper=None,
                       lower_constraints=[], upper_constraints=[],
                       rgen=None):
        """Add a time parameter to the demographic model.

        :param str name: Parameter name
        :param float t0: Starting value. If None, use ``rgen`` to randomly sample
        :param float lower: Lower bound
        :param float upper: Upper bound
        :param list lower_constraints: List of parameter names that \
        are lower bounds
        :param list upper_constraints: List of parameter names that \
        are upper bounds
        :param rgen: Function to sample a random starting value. If \
        None, a truncated exponential with rate \
        ``1 / (N_e * gen_time)`` constrained to satisfy the bounds \
        and constraints.
        """
        def lower_bound(params):
            constraints = [params[k] for k in lower_constraints]
            constraints.append(lower)
            return np.max(np.array(constraints))

        def upper_bound(params):
            constraints = [params[k] for k in upper_constraints]
            if upper is not None:
                constraints.append(upper)
            return np.min(np.array(constraints))

        has_upper = (len(upper_constraints) > 0) or (upper is not None)

        def scale_transform(t, params):
            l = lower_bound(params)
            if t < l:
                raise ValueError(
                    "t = {} < {} = max({})".format(
                        t, l, [lower] + list(lower_constraints)))
            if has_upper:
                u = upper_bound(params)
                if t > u:
                    raise ValueError(
                        "t = {} > {} = min({})".format(
                            t, u, [upper] + list(upper_constraints)))

                x = (t - l) / (u - l)
                x = np.log(x/(1.-x))
                return x
            else:
                x = t - l
                return x

        def unscale_transform(x, params):
            l = lower_bound(params)
            if has_upper:
                x = 1./(1.+np.exp(-x))
                u = upper_bound(params)
                return (1-x)*l + x*u
            else:
                return l + x

        if rgen is None:
            scale = self.N_e * self.gen_time  # average time to coalescence

            def rgen(params):
                l = lower_bound(params)
                if has_upper:
                    u = upper_bound(params)
                    b = (u-l)/scale
                else:
                    b = float("inf")
                truncexpon = scipy.stats.truncexpon(b=b, loc=l, scale=scale)
                return truncexpon.rvs()

        if has_upper:
            self.add_parameter(
                name, t0,
                scale_transform=scale_transform,
                unscale_transform=unscale_transform,
                rgen=rgen)
        else:
            self.add_parameter(
                name, t0,
                scaled_lower=1e-12,  # prevent optimizer from having a slightly negative number
                scale_transform=scale_transform,
                unscale_transform=unscale_transform,
                rgen=rgen)

    def add_pulse_param(self, name, p0=None, lower=0.0, upper=1.0, rgen=None):
        """Add a pulse parameter to the demographic model.

        :param str name: Parameter name.
        :param float p0: Starting value. If None, randomly sample \
        with ``rgen``
        :param float lower: Lower bound
        :param float upper: Upper bound
        :param function rgen: Function to sample random value. \
        If None, use a uniform distribution.
        """
        if rgen is None:
            def rgen(params):
                return np.random.uniform(lower, upper)

        def scale_transform(x, params):
            return np.log(x/np.array(1-x))

        if lower == 0:
            # avoid log(0)
            scaled_lower = None
        else:
            scaled_lower = scale_transform(lower, None)

        if upper == 1:
            # avoid divide by 0 warning
            scaled_upper = None
        else:
            scaled_upper = scale_transform(upper, None)

        self.add_parameter(name, p0,
                           scaled_lower=scaled_lower,
                           scaled_upper=scaled_upper,
                           scale_transform=scale_transform,
                           unscale_transform=lambda x, params: 1/(1+np.exp(-x)),
                           rgen=rgen)

    def add_growth_param(self, name, g0=None, lower=-.001, upper=.001,
                         rgen=None):
        """Add growth rate parameter to the demographic model.

        :param str name: Parameter name
        :param float g0: Starting value. If None, randomly sample \
        with ``rgen``
        :param float lower: Lower bound
        :param float upper: Upper bound
        :param function rgen: Function to sample random value. \
        If None, use uniform distribution
        """
        if rgen is None:
            def rgen(params):
                return np.random.uniform(lower, upper)
        self.add_parameter(name, g0, scaled_lower=lower, scaled_upper=upper,
                           rgen=rgen)

    def add_leaf(self, pop_name, t=0, N=None, g=None):
        """Add a sampled leaf population to the model.

        The arguments t, N, g can be floats or parameter names (strings).

        If N or g are specified, then the size or growth rate are also set at
        the sampling time t. Otherwise they remain at their previous (default)
        values.

        Note that this does not affect the population size and growth below
        time t, which may be an issue if lineages are moved in from other
        populations below t. If you need to set the size or growth below t, use
        :meth:`DemographicModel.set_size()`.

        :param str pop_name: Name of the population
        :param float,str t: Time the population was sampled
        :param float,str N: Population size
        :param float,str g: Population growth rate

        """
        self.leafs.append(pop_name)

        self.leaf_events.append(LeafEvent(
            t, pop_name, self.N_e, self.gen_time))

        if N is not None:
            self.set_size(pop_name, t, N=N)
        if g is not None:
            self.set_size(pop_name, t, g=g)

    def move_lineages(self, pop_from, pop_to, t, p=1, N=None, g=None):
        """Move each lineage in pop_from to pop_to at time t with probability p.

        The arguments t, p, N, g can be floats or parameter names (strings).

        If N or g are specified, then the size or growth rate of pop_to is also
        set at this time, otherwise these parameters remain at their previous
        values.

        :param str pop_from: Population lineages are moved from \
        (backwards in time)
        :param str pop_to: Population lineages are moved to \
        (backwards in time)
        :param float,str t: Time of the event
        :param float,str p: Probability that lineage in pop_from \
        moves to pop_to
        :param float,str N: Population size of pop_to
        :param float,str g: Growth rate of pop_to
        """
        if p == 1:
            self.topology_events.append((JoinEvent(
                t, pop_from, pop_to, self.N_e, self.gen_time)))
        else:
            self.topology_events.append(PulseEvent(
                t, p, pop_from, pop_to, self.N_e,
                self.gen_time))

        if N is not None:
            self.set_size(pop_to, t, N=N)
        if g is not None:
            self.set_size(pop_to, t, g=g)

    def set_size(self, pop_name, t, N=None, g=0):
        """Set population size and/or growth rate at time t.

        The arguments t, N, g can be floats or parameter names (strings).

        If N is not specified then only the growth rate is changed.

        If N is specified and g is not then the growth rate is reset to 0.
        Currently it is not possible to change the size without also setting
        the growth rate.

        :param str pop_name: Population name
        :param float,str t: Time of event
        :param float,str N: Population size
        :param float,str g: Growth rate
        """
        if N is not None:
            self.size_events.append(SizeEvent(
                t, N, pop_name, self.N_e, self.gen_time))
        self.size_events.append(GrowthEvent(
            t, g, pop_name, self.N_e, self.gen_time))

    def get_params(self, scaled=False):
        """Return an ordered dictionary with the current parameter
        values.

        If ``scaled=True``, returns the parameters scaled in the internal
        representation used by momi during optimization (see also:
        :meth:`DemographicModel.add_parameter`)
        """
        params_dict = ParamsDict()
        for param in self.parameters.values():
            if scaled:
                params_dict[param.name] = param.x
            else:
                param.update_params_dict(params_dict)
        return params_dict

    def _get_x(self, param=None):
        if param is None:
            return np.array([
                p.x for p in self.parameters.values()])
        else:
            for p in self.parameters.values():
                if p.name == param:
                    return p.x
            raise ValueError("Unrecognized parameter {}".format(param))

    def _set_x(self, x, param=None):
        if param:
            x = {param: x}
        self.set_params(x, scaled=True)

    def simulate_data(self, length, recoms_per_gen,
                      num_replicates, muts_per_gen=None,
                      sampled_n_dict=None, **kwargs):
        """Simulate data, using msprime as backend

        :param int length: Length of each locus in bases
        :param float recoms_per_gen: Recombination rate per generation per base
        :param float muts_per_gen: Mutation rate per generation per base
        :param int num_replicates: Number of loci to simulate
        :param dict sampled_n_dict: Number of haploids per population. \
        If None, use sample sizes from the current dataset as set by \
        :meth:`DemographicModel.set_data`

        :returns: Dataset of SNP allele counts
        :rtype: :class:`SnpAlleleCounts`
        """
        demo = self._get_demo(sampled_n_dict)
        if muts_per_gen is None:
            if not self.muts_per_gen:
                raise ValueError("Need to provide mutation rate")
            muts_per_gen = self.muts_per_gen
        return demo.simulate_data(
            length=length,
            recombination_rate=4*self.N_e*recoms_per_gen,
            mutation_rate=4*self.N_e*muts_per_gen,
            num_replicates=num_replicates,
            **kwargs)

    def simulate_vcf(
            self, out_prefix,
            recoms_per_gen, length,
            muts_per_gen=None, chrom_name="1",
            ploidy=1, random_seed=None,
            sampled_n_dict=None, **kwargs):
        """Simulate a chromosome using msprime and write it to VCF

        :param str,file outfile: Output VCF file. If a string ending in ".gz", gzip it.
        :param float muts_per_gen: Mutation rate per generation per base
        :param float recoms_per_gen: Recombination rate per generation per base
        :param int length: Length of chromosome in bases
        :param str chrom_name: Name of chromosome
        :param int ploidy: Ploidy
        :param int random_seed: Random seed
        :param dict sampled_n_dict: Number of haploids per population. \
        If None, use sample sizes from the current dataset as set by \
        :meth:`DemographicModel.set_data`
        """
        demo = self._get_demo(sampled_n_dict)
        if muts_per_gen is None:
            if not self.muts_per_gen:
                raise ValueError("Need to provide mutation rate")
            muts_per_gen = self.muts_per_gen
        return demo.simulate_vcf(
            out_prefix=out_prefix,
            mutation_rate=4*self.N_e*muts_per_gen,
            recombination_rate=4*self.N_e*recoms_per_gen,
            length=length, chrom_name=chrom_name,
            ploidy=ploidy, random_seed=random_seed,
            **kwargs)

    # TODO rename to get_multipop_moran?
    def _get_demo(self, sampled_n_dict):
        sampled_n_dict = self._get_sample_sizes(sampled_n_dict)

        params_dict = self.get_params()

        events = []
        for sub_events in (self.leaf_events,
                           self.size_events,
                           self.topology_events):
            for e in sub_events:
                events.append(e)

        events = sorted(events, key=lambda e: e.t(params_dict))
        G = _build_demo_graph(events, sampled_n_dict, params_dict, default_N=1.0)
        demo = Demography(G)

        def printable_params():
            for k, v in params_dict.items():
                v = str(v).replace('\n', '')
                yield (k, v)
        logging.getLogger(__name__).debug("Demographic parameters = {}".format(
            co.OrderedDict(printable_params())))

        return demo

    def _demo_fun(self, *x):
        logging.getLogger(__name__).debug("x = {}".format(str(x)))
        prev_x = self._get_x()
        try:
            self._set_x(x)
            return self._get_demo(None)
        except Exception as e:
            curr_params = self.get_params()
            curr_scaled = self.get_params(scaled=True)

            curr_params = {k: str(v) for k, v in curr_params.items()}
            curr_scaled = {k: str(v) for k, v in curr_scaled.items()}

            new_msg = "Exception encountered at parameter values {} (internal scaling x = {})".format(curr_params, curr_scaled)
            raise ValueError(new_msg) from e
        finally:
            self._set_x(prev_x)

    def set_data(
            self, sfs, length=None,
            mem_chunk_size=1000,
            non_ascertained_pops=None,
            use_pairwise_diffs=True):
        """Set dataset for the model.

        :param Sfs sfs: Observed SFS
        :param float length: Length of data in bases. Overrides ``sfs.length`` if set. Required if :attr:`DemoModel.muts_per_gen` is set and ``sfs.length`` is not.
        :param mem_chunk_size: Controls memory usage by computing likelihood in chunks of SNPs. If ``-1`` then no chunking is done.
        :param non_ascertained_pops: Don't ascertain SNPs within these populations. That is, ignore all SNPs that are not polymorphic on the other populations. The SFS is adjusted to represent probabilities conditional on this ascertainment scheme.
        :param use_pairwise_diffs: Only has an effect if :attr:`DemoModel.muts_per_gen` is set. If ``False``, assumes the total number of mutations is Poisson. If True, models the within population nucleotide diversity (i.e. the average number of heterozygotes per population) as independent Poissons. If there is missing data this is required to be ``True``.
        """
        if not length:
            length = sfs.length

        self._set_data(
            sfs=sfs, length=length,
            mem_chunk_size=mem_chunk_size,
            use_pairwise_diffs=use_pairwise_diffs,
            non_ascertained_pops=non_ascertained_pops)

    def _set_data(self, sfs, length,
                  mem_chunk_size, use_pairwise_diffs,
                  non_ascertained_pops):
        self._lik_surface = None
        self._conf_region = None
        self._subsfs = None
        self._fullsfs = sfs
        self._length = length
        self._mem_chunk_size = mem_chunk_size
        self._use_pairwise_diffs = use_pairwise_diffs
        self._non_ascertained_pops = non_ascertained_pops

    def _get_sfs(self):
        if self._subsfs is None or list(
                self._subsfs.sampled_pops) != list(self.leafs):
            self._subsfs = self._fullsfs.subset_populations(
                self.leafs,
                non_ascertained_pops=self._non_ascertained_pops).sfs
        return self._subsfs

    def fit_within_pop_diversity(self):
        """Estimates mutation rate using within-population nucleotide diversity.

        The within-population nucleotide diversity is the average number of hets \
        per individual in the population, assuming it is at Hardy Weinberg equilibrium.

        This returns an estimated mutation rate for each (ascertained) population. Note these are \
        non-independent estimates of the same value. It also returns standard deviations from the jacknife.

        If :attr:`DemographicModel.muts_per_gen` is set, will also return Z-scores \
        of the residuals.

        :rtype: :class:`pandas.DataFrame`
        """
        if not self._length:
            raise ValueError("SFS has no length attribute, need to provide it with set_data(). If unknown, set length=1 to estimate the mutation rate over the full data instead of per-base.")

        sfs = self._get_sfs()

        keep_pops = np.array([
            a and n > 1
            for p, n, a in zip(sfs.sampled_pops, sfs.sampled_n,
                               sfs.ascertainment_pop)])
        asc_pops = np.array(sfs.sampled_pops)[keep_pops]

        # expected hets per unit mutation
        e_het = expected_heterozygosity(
            self._get_demo({p: 2 for p in sfs.sampled_pops}),
            restrict_to_pops=asc_pops)
        # convert from ms scaling
        e_het = e_het * 4 * self.N_e

        # scale by data length and per-population missingness
        p_miss = self._p_missing_dict()
        p_miss = np.array([p_miss[pop] for pop in asc_pops])
        e_het = e_het * self._length * (1-p_miss)

        # observed hets per locus
        o_het = sfs.avg_pairwise_hets[:, keep_pops]
        # sum over loci
        total_o_het = o_het.sum(axis=0)

        # estimated mutation rate per population
        est_mut_rate = total_o_het / e_het
        # jaccknife
        n_loci = o_het.shape[0]
        jackknife_o_het = (total_o_het - o_het) * (n_loci / (n_loci-1))
        jackknife_mut_rate = jackknife_o_het / e_het

        # make dataframe
        if not self.muts_per_gen:
            muts_per_gen = 0
        else:
            muts_per_gen = self.muts_per_gen

        jackknife_stats = []
        for i, pop in enumerate(asc_pops):
            jackknife_stats.append(JackknifeGoodnessFitStat(
                muts_per_gen, est_mut_rate[i],
                jackknife_mut_rate[:, i]))

        df = pd.DataFrame(
            [[p, j.observed, j.sd]
             for p, j in zip(asc_pops, jackknife_stats)],
            columns=["Pop", "EstMutRate", "JackknifeSD"])
        if muts_per_gen:
            df["JackknifeZscore"] = [j.z_score
                                     for j in jackknife_stats]
        return df

    #def _get_conf_region(self):
    #    opt_surface = self._get_surface()
    #    opt_x = self._get_opt_x()
    #    if self._conf_region is None or not np.allclose(
    #            opt_x, self._conf_region.point):
    #        opt_score = opt_surface._score(opt_x)
    #        opt_score_cov = opt_surface._score_cov(opt_x)
    #        opt_fisher = opt_surface._fisher(opt_x)

    #        self._conf_region = _ConfidenceRegion(
    #            opt_x, opt_score, opt_score_cov, opt_fisher,
    #            lik_fun=opt_surface.log_lik,
    #            psd_rtol=1e-4)
    #    return self._conf_region

    #def marginal_wald(self):
    #    marginal_wald_df = co.OrderedDict()
    #    marginal_wald_df["Param"] = [
    #        p.name for p in self.parameters.values()]
    #    marginal_wald_df["Value"] = list(
    #        self.get_params().values())
    #    marginal_wald_df["StdDev"] = np.sqrt(np.diag(self.mle_cov()))
    #    return pd.DataFrame(marginal_wald_df)

    #def mle_cov(self, x_scale=False):
    #    # use delta method
    #    G = self._get_conf_region().godambe(inverse=True)
    #    if x_scale:
    #        return G
    #    else:
    #        dp_do = self._get_params_opt_x_jacobian()
    #        return np.dot(dp_do, np.dot(G, dp_do.T))

    #def test(self, null_point=None, sims=int(1e3), test_type="ratio",
    #         alt_point=None, *args, **kwargs):
    #    if null_point is None:
    #        null_point = self._get_x()
    #    null_point = self._opt_x_from_x(null_point)
    #    if alt_point is not None:
    #        alt_point = self._opt_x_from_x(alt_point)
    #    return self._get_conf_region().test(
    #        null_point=null_point, sims=sims,
    #        test_type=test_type, alt_point=alt_point, *args, **kwargs)

    def _get_surface(self):
        if self._lik_surface is not None and (
                list(self._lik_surface.data.sampled_pops) ==
                list(self.leafs)):
            return self._lik_surface

        self._conf_region = None
        if self._fullsfs is None:
            raise ValueError("Need to call DemographicModel.set_data()")
        # TODO better message (e.g. "Building SFS...")
        logging.getLogger(__name__).info("Constructing likelihood surface...")

        sfs = self._get_sfs().combine_loci()
        use_pairwise_diffs = self._use_pairwise_diffs

        muts_per_gen = self.muts_per_gen
        if muts_per_gen is None:
            mut_rate = None
        elif self._length is None:
            raise ValueError("SFS is missing length attribute, need to manually set length in set_data(), or set mutation rate to None")
        else:
            mut_rate = 4 * self.N_e * muts_per_gen * self._length

        demo_fun = self._demo_fun

        p_miss = self._p_missing_dict()
        p_miss = np.array([p_miss[pop] for pop in sfs.sampled_pops])

        self._lik_surface = SfsLikelihoodSurface(
            sfs, demo_fun, mut_rate=mut_rate,
            folded=sfs.folded, batch_size=self._mem_chunk_size,
            use_pairwise_diffs=use_pairwise_diffs, p_missing=p_miss)

        logging.getLogger(__name__).info("Finished constructing likelihood surface")

        return self._lik_surface

    def _p_missing_dict(self):
        p_miss = self._fullsfs.p_missing
        return {pop: pm for pop, pm in zip(
            self._fullsfs.populations, p_miss)}

    # NOTE note these are in PER-GENERATION units
    # TODO allow to pass folded parameter (if passing separate configs?)
    def expected_sfs(self, configs=None, normalized=False,
                     folded=False, length=None,
                     return_dict=True):
        if configs is None:
            sfs = self._get_sfs()
            configs = sfs.configs
            folded = sfs.folded

        demo = self._get_demo(dict(zip(configs.sampled_pops,
                                       configs.sampled_n)))
        ret = expected_sfs(demo, configs,
                           normalized=normalized, folded=folded)
        if not normalized:
            if length is None:
                length = self._length
            ret = ret * self.N_e * 4.0 * self.muts_per_gen * length
        if return_dict:
            return co.OrderedDict(zip(configs.as_tuple(), ret))
        else:
            return ret

    # NOTE these are in PER-GENERATION units
    # this is mainly here for some old unit tests...
    def expected_branchlen(self, sampled_n_dict=None):
        demo = self._get_demo(sampled_n_dict)
        return expected_total_branch_len(demo) * self.N_e * 4.0

    def _get_sample_sizes(self, sampled_n_dict):
        if sampled_n_dict is not None:
            sampled_pops_set = set(sampled_n_dict.keys())
            leaf_set = set(self.leafs)
            if not sampled_pops_set <= leaf_set:
                raise ValueError("{} not in leaf populations".format(
                    sampled_pops_set - leaf_set))
            # make sure it is sorted in correct order
            return co.OrderedDict((k, sampled_n_dict[k]) for k in self.leafs
                                  if k in sampled_pops_set)
        if self._fullsfs is None:
            raise ValueError(
                "Need to call set_data() or provide dict of sample sizes.")
        sfs = self._get_sfs()
        return co.OrderedDict(zip(sfs.sampled_pops, sfs.sampled_n))

    def log_likelihood(self):
        """
        The log likelihood at the current parameter values
        """
        return self._get_surface().log_lik(self._get_x())

    def kl_div(self):
        """
        The KL-divergence at the current parameter values
        """
        return self._get_surface().kl_div(self._get_x())

    def stochastic_optimize(
            self, num_iters, n_minibatches=None, snps_per_minibatch=None,
            rgen=None, printfreq=1, start_from_checkpoint=None,
            save_to_checkpoint=None,  svrg_epoch=-1, **kwargs):
        """Use stochastic optimization (ADAM+SVRG) to search for MLE

        Exactly one of of ``n_minibatches`` and ``snps_per_minibatch`` should be set, as one determines the other.

        :param int num_iters: Number of steps
        :param int n_minibatches: Number of minibatches
        :param int snps_per_minibatch: Number of SNPs per minibatch
        :param numpy.RandomState rgen: Random generator
        :param int printfreq: How often to log progress
        :param str start_from_checkpoint: Name of checkpoint file to start from
        :param str save_to_checkpoint: Name of checkpoint file to save to
        :param int svrg_epoch: How often to compute full likelihood for SVRG. -1=never.
        :rtype: :class:`scipy.optimize.OptimizeResult`
        """
        def callback(x):
            self._set_x(x)
            if x.iteration % printfreq == 0:
                msg = [("it", x.iteration), ("LogLikelihood", -x.fun)]
                msg.extend(list(self.get_params().items()))
                msg = ", ".join(["{}: {}".format(k, v) for k, v in msg])
                logging.getLogger(__name__).info("{" + msg + "}")

        if rgen is None:
            rgen = np.random

        bounds = [p.x_bounds
                  for p in self.parameters.values()]
        if all([b is None for bnd in bounds for b in bnd]):
            bounds = None

        kwargs = dict(kwargs)
        kwargs["callback"] = callback
        kwargs["bounds"] = bounds

        if start_from_checkpoint:
            with open(start_from_checkpoint) as f:
                kwargs.update(json.load(f))
        else:
            kwargs["x0"] = self._get_x()

        res = self._get_surface()._stochastic_surfaces(
            n_minibatches=n_minibatches,
            snps_per_minibatch=snps_per_minibatch,
            rgen=rgen).find_mle(
                method="adam", num_iters=num_iters,
                svrg_epoch=svrg_epoch,
                checkpoint_file=save_to_checkpoint, **kwargs)

        self._set_x(res.x)
        res["parameters"] = self.get_params()
        res["cross_entropy"] = res.fun
        return res

    def optimize(self, method="tnc", jac=True,
                 hess=False, hessp=False, printfreq=1, **kwargs):
        """Search for the maximum likelihood value of the parameters.

        This is a wrapper around :func:`scipy.optimize.minimize`, \
        and arguments for that function can be passed in via \
        ``**kwargs``. Note the following arguments are constructed by :mod:`momi` and not be passed in by ``**kwargs``: ``fun``, ``x0``, ``jac``, ``hess``, ``hessp``, ``bounds``.

        :param str method: Optimization method. Default is "tnc". For large models "L-BFGS-B" is recommended. See :func:`scipy.optimize.minimize`.
        :param bool jac: Whether or not to provide the gradient (computed via :mod:`autograd`) to the optimizer. If `False`, optimizers requiring gradients will typically approximate it via finite differences.
        :param bool hess: Whether or not to provide the hessian (computed via :mod:`autograd`) to the optimizer.
        :param bool hessp: Whether or not to provide the hessian-vector-product (via :mod:`autograd`) to the optimizer
        :param int printfreq: Log current progress via :func:`logging.info` every `printfreq` iterations
        :rtype: :class:`scipy.optimize.OptimizeResult`
        """
        bounds = [p.x_bounds
                  for p in self.parameters.values()]
        if all([b is None for bnd in bounds for b in bnd]):
            bounds = None

        def callback(x):
            self._set_x(x)
            if x.iteration % printfreq == 0:
                msg = [("it", x.iteration), ("KLDivergence", x.fun)]
                msg.extend(list(self.get_params().items()))
                msg = ", ".join(["{}: {}".format(k, v) for k, v in msg])
                logging.getLogger(__name__).info("{" + msg + "}")

        res = self._get_surface().find_mle(
            self._get_x(), method=method,
            jac=jac, hess=hess, hessp=hessp,
            bounds=bounds, callback=callback,
            **kwargs)

        self._set_x(res.x)
        res["parameters"] = self.get_params()
        res["kl_divergence"] = res.fun
        res["log_likelihood"] = self.log_likelihood()
        return res



import collections as co
import functools as ft
from cached_property import cached_property
import autograd.numpy as np
from .compute_sfs import _expected_sfs_tensor_prod

import pandas as pd
import seaborn
from matplotlib import pyplot as plt


class SfsStats(object):
    def __init__(self, sampled_n_dict):
        self.sampled_n_dict = {p: n for p, n in sampled_n_dict.items()
                               if n > 0}

    def tensor_prod(self, derived_weights_dict):
        raise NotImplementedError

    def log(self, x):
        raise NotImplementedError

    @cached_property
    def denom(self):
        return self.tensor_prod({})

    def ordered_prob(self, subsample_dict,
                     fold=False):
        # The ordered probability for the subsample given by
        # subsample_dict.
        #
        # Parameters:
        # subsample_dict: dict of list
        #    dict mapping population to a list of 0s and 1s giving the
        #       ordered subsample within that population.

        if fold:
            rev_subsample = {p: 1 - np.array(s)
                             for p, s in subsample_dict.items()}

            return (self.ordered_prob(subsample_dict)
                    + self.ordered_prob(rev_subsample))

        derived_weights_dict = {}
        for pop, pop_subsample in subsample_dict.items():
            n = self.sampled_n_dict[pop]
            arange = np.arange(n+1)

            cnts = co.Counter(pop_subsample)

            prob = np.ones(n+1)
            for i in range(cnts[0]):
                prob *= (n - arange - i)
            for i in range(cnts[1]):
                prob *= (arange - i)
            for i in range(cnts[0] + cnts[1]):
                prob /= float(n - i)

            derived_weights_dict[pop] = prob

        return self.tensor_prod(derived_weights_dict) / self.denom

    def count_1100(self, A, B, C, O=None):
        # O=None -> O is the root population
        subsample_dict = co.defaultdict(list)
        subsample_dict[A].append(1)
        subsample_dict[B].append(1)
        subsample_dict[C].append(0)
        if O is not None:
            subsample_dict[O].append(0)

        return self.ordered_prob(
            subsample_dict,
            fold=(O is not None))

    def baba(self, A, B, C, *O):
        return self.count_1100(A, C, B, *O)

    def abba(self, A, B, C, *O):
        return self.count_1100(B, C, A, *O)

    def abba_baba(self, A, B, C, D=None):
        """
        Same as :meth:`f4`
        """
        return self.baba(A, B, C, D) - self.abba(A, B, C, D)

    def log_abba_baba(self, A, B, C, D=None):
        """
        Returns log(BABA/ABBA) = log(BABA)-log(ABBA)
        """
        return self.log(self.baba(A, B, C, D)) - self.log(self.abba(A, B, C, D))

    def f_st(self, A, B):
        """
        Returns (pi_between - pi_within) / pi_between, \
        where pi_between, pi_within represent the average number of pairwise \
        diffs between 2 individuals sampled from different or the same \
        population, respectively.
        """
        pi_within = (self.ordered_prob({A: [1, 0]}, fold=True)
                     + self.ordered_prob({B: [1, 0]}, fold=True)) / 2
        pi_between = self.ordered_prob({A: [1], B: [0]}, fold=True)

        return (pi_between - pi_within) / pi_between


    def f4(self, A, B, C, D=None):
        """
        Returns the ABBA-BABA (f4) statistic for testing admixture.

        :param str A: First population
        :param str B: Second population
        :param str C: Third population
        :param str D: Fourth population. If None, use ancestral allele.
        """
        return self.abba_baba(A, B, C, D)

    def f3(self, A, B, O):
        """
        Computes f3 statistic (O-A)*(O-B)

        :param str A: First population
        :param str B: Second population
        :param str O: Third population.
        """
        return self.f4(O, A, O, B)

    def f2(self, A, B):
        """
        Computes f2 statistic (A-B)*(A-B)

        :param str A: First population
        :param str B: Second population
        """
        return self.f4(A, B, A, B)

    def pattersons_d(self, A, B, C, D=None):
        """
        Returns Patterson's D, defined as (BABA-ABBA)/(BABA+ABBA).

        :param str A: First population
        :param str B: Second population
        :param str C: Third population
        :param str D: Fourth population. If None, use ancestral allele.
        """
        abba = self.abba(A, B, C, D)
        baba = self.baba(A, B, C, D)
        return (baba - abba) / (baba + abba)

    def greens_f(self, A, B, C, *O):
        # Estimate for the admixture of C into B in tree (((A,B),C),O)
        return self.abba_baba(A, B, C, *O) / self.abba_baba(A, C, C, *O)

    def f4_ratio(self, A, B, C, X, *O):
        # For tree (((A,B),C),O), and X admixed between B,C
        # an estimate for the admixture proportion from B
        # ref: Patterson et al 2012, Ancient Admixture in Human History, eq (4)
        return self.f4(X, C, A, *O) / self.f4(B, C, A, *O)

    def singleton_probs(self, pops):
        denom = None
        probs = {}
        for pop in pops:
            prob = self.ordered_prob(dict([
                (p, [1]) if p == pop else (p, [0])
                for p in pops]), fold=True)
            probs[pop] = prob
            if denom is None:
                denom = prob
            else:
                denom = denom + prob
        return {"probs": probs, "denom": 1-denom}


class SfsModelFitStats(SfsStats):
    """Class to compare expected vs. observed statistics of the SFS.

    All methods return :class:`JackknifeGoodnessFitStat` unless
    otherwise stated.

    Currently, all goodness-of-fit statistics are based on the multinomial
    SFS (i.e., the SFS normalized to be a probability distribution
    summing to 1). Thus the mutation rate has no effect on these statistics.

    See `Patterson et al 2012 <http://www.genetics.org/content/192/3/1065>`_,
    for definitions of f2, f3, f4 (abba-baba), and D statistics.

    Note this class does NOT get updated when the underlying
    ``demo_model`` changes; a new :class:`SfsModelFitStats` needs
    to be created to reflect any changes in the demography.

    :param momi.DemographicModel demo_model: Demography to compute expected \
    statistics under.

    :param dict sampled_n_dict: The number of samples to use \
    per population. SNPs with fewer than this number of samples \
    are ignored. The default is to use the \
    full sample size of the data, i.e. to remove all SNPs with any missing \
    data. For datasets with large amounts of missing data, \
    this could potentially lead to most or all SNPs being removed, so it is \
    important to specify a smaller sample size in such cases.
    """
    def __init__(self, demo_model, sampled_n_dict=None):
        if not demo_model._fullsfs:
            raise ValueError(
                "Need to call DemographicModel.set_data() first")

        sfs = demo_model._get_sfs()
        if not sampled_n_dict:
            sampled_n_dict = dict(zip(sfs.sampled_pops, sfs.sampled_n))
        self.sampled_n_dict = sampled_n_dict

        self.empirical = ObservedSfsStats(sfs, self.sampled_n_dict)
        self.expected = ExpectedSfsStats(
            demo_model._get_demo(self.sampled_n_dict), [
                pop for pop, is_asc in zip(sfs.sampled_pops,
                                           sfs.ascertainment_pop)
                if is_asc])

    def tensor_prod(self, derived_weights_dict):
        """Compute rank-1 tensor products of the SFS, which can be used \
        to express a wide range of SFS-based statistics.

        More specifically, this computes the sum

        .. math:: \sum_{i,j,\ldots} SFS_{i,j,\ldots} w^{(1)}_i w^{(2)}_j \cdots

        where :math:`w^{(1)}_i` is the weight corresponding to SFS entries \
        with ``i`` derived alleles in population 1, etc. Note the SFS is \
        normalized to sum to 1 here (it is a probability).

        :param dict derived_weights_dict: Maps leaf populations to \
        vectors (:class:`numpy.ndarray`). If a population has ``n`` samples \
        then the corresponding vector ``w`` should have length ``n+1``, \
        with ``w[i]`` being the weight for SFS entries with ``i`` copies of \
        the derived allele in the population.

        :rtype: :class:`JackknifeGoodnessFitStat`
        """
        exp = self.expected.tensor_prod(derived_weights_dict)
        emp = self.empirical.tensor_prod(derived_weights_dict)

        exp = exp / self.expected.denom
        emp = emp / self.empirical.denom

        return JackknifeGoodnessFitStat(exp, emp.est, emp.jackknife)

    def log(self, x):
        return x.apply(np.log)

    @property
    def denom(self):
        return 1.0

    def all_pairs_ibs(self, fig=True):
        """Fit the IBS fraction for all pairs of populations, and optionally plot it.

        :param bool fig: whether to plot it
        :rtype: :class:`pandas.DataFrame`
        """
        pops = list(self.sampled_n_dict.keys())

        df = []
        for pop1 in pops:
            for pop2 in pops:
                if pop1 > pop2:
                    continue
                elif pop1 == pop2:
                    if self.sampled_n_dict[pop1] == 1:
                        continue
                    prob = self.ordered_prob({
                        pop1: [0, 0]}, fold=True)
                else:
                    prob = self.ordered_prob({
                        pop1: [0], pop2: [0]}, fold=True)

                line = [pop1, pop2, prob.expected,
                        prob.observed, prob.z_score]
                df.append(line)

        return self._pairwise_zscores(df, fig)

    def all_f2(self, fig=True):
        pops = [k for k, v in self.sampled_n_dict.items() if v > 1]

        df = []
        for pop1 in pops:
            for pop2 in pops:
                if pop1 >= pop2:
                    continue
                else:
                    prob = self.f2(pop1, pop2)

                line = [pop1, pop2, prob.expected,
                        prob.observed, prob.z_score]
                df.append(line)

        return self._pairwise_zscores(df, fig)

    def _pairwise_zscores(self, df, fig):
        ret = pd.DataFrame(sorted(df, key=lambda x: abs(x[-1]),
                                  reverse=True),
                           columns=["Pop1", "Pop2",
                                    "Expected", "Observed", "Z"])

        if fig:
            pivoted = ret.pivot(index="Pop1", columns="Pop2",
                                values="Z")
            plt.gcf().clear()
            seaborn.heatmap(pivoted, annot=True, center=0)
            plt.title("Residual (Observed-Expected) Z-scores")
            pass
        return ret

    @property
    def n_subsets(self):
        return self.empirical.n_subsets

    @property
    def n_jackknife_blocks(self):
        return self.empirical.n_jackknife_blocks


class ObservedSfsStats(SfsStats):
    def __init__(self, sfs, sampled_n_dict):
        is_ascertained = dict(zip(sfs.sampled_pops, sfs.ascertainment_pop))
        if sum(n for p, n in sampled_n_dict.items()
               if is_ascertained[p]) < 2:
            raise ValueError("sampled_n_dict must contain at least 2 ascertained alleles")
        self.sfs = sfs
        super(ObservedSfsStats, self).__init__(sampled_n_dict)

    def tensor_prod(self, derived_weights_dict):
        weighted_counts = self.sfs.configs.count_subsets(derived_weights_dict,
                                                         self.sampled_n_dict)

        # subtract out weights of monomorphic
        mono_anc = {}
        mono_der = {}
        for pop, asc in zip(self.sfs.sampled_pops, self.sfs.ascertainment_pop):
            try:
                v = derived_weights_dict[pop]
            except KeyError:
                try:
                    v = [1] * (self.sampled_n_dict[pop]+1)
                except KeyError:
                    continue
            if asc:
                mono_anc[pop] = [v[0]] + [0]*(len(v)-1)
                mono_der[pop] = [0]*(len(v)-1) + [v[-1]]
            else:
                mono_anc[pop] = v
                mono_der[pop] = v
        mono_anc = self.sfs.configs.count_subsets(
            mono_anc, self.sampled_n_dict)
        mono_der = self.sfs.configs.count_subsets(
            mono_der, self.sampled_n_dict)

        return JackknifeStat.from_chunks(
            self.sfs.freqs_matrix.T.dot(
                weighted_counts - mono_anc - mono_der))

    def log(self, x):
        return x.apply(np.log)

    @property
    def n_subsets(self):
        return self.denom.est

    @property
    def n_jackknife_blocks(self):
        return self.sfs.n_loci


class ExpectedSfsStats(SfsStats):
    def __init__(self, demo, ascertainment_pops):
        self.demo = demo
        self.ascertainment_pops = ascertainment_pops
        super(ExpectedSfsStats, self).__init__(dict(zip(demo.sampled_pops,
                                                        demo.sampled_n)))

    def tensor_prod(self, derived_weights_dict):
        #sampled_pops, sampled_n = zip(*sorted(self.sampled_n_dict.items()))
        #demo = self.demo._get_multipop_moran(sampled_pops, sampled_n)
        demo = self.demo

        vecs = []
        for p, n in zip(demo.sampled_pops, demo.sampled_n):
            v = []
            try:
                row = derived_weights_dict[p]
            except KeyError:
                row = np.ones(n+1)
            assert len(row) == n+1

            if p in self.ascertainment_pops:
                v.append([row[0]] + [0.0] * n)  # all ancestral state
                v.append([0.0] * n + [row[-1]])  # all derived state
            else:
                for _ in range(2):
                    v.append(row)
            v.append(row)

            vecs.append(np.array(v))

        res = _expected_sfs_tensor_prod(vecs, demo)
        return res[2] - res[0] - res[1]

    def log(self, x):
        return np.log(x)


class JackknifeGoodnessFitStat(object):
    """
    Object returned by methods of :class:`SfsModelFitStats`.

    Basic arithmetic operations are supported, allowing to build
    up complex statistics out of simpler ones.

    The raw expected, observed, and jackknifed_array values
    can be accessed as attributes of this class.

    :param float expected: the expected value of the statistic
    :param float observed: the observed value of the statistic
    :param numpy.ndarray jackknifed_array: array of the jackknifed \
    values of the statistic.
    """
    def __init__(self, expected, observed, jackknifed_array):
        self.expected = expected
        self.observed = observed
        self.jackknifed_array = jackknifed_array

    @property
    def sd(self):
        """
        Standard deviation of the statistic, estimated via jackknife
        """
        resids = self.jackknifed_array - self.observed
        return np.sqrt(np.mean(resids**2) * (
            len(self.jackknifed_array) - 1))

    @property
    def bias(self):
        return np.mean(self.jackknifed_array - self.observed) * (
            len(self.jackknifed_array) - 1)

    @property
    def z_score(self):
        """
        Z-score of the statistic, defined as (observed-expected)/sd
        """
        return (self.observed - self.expected) / self.sd

    def __repr__(self):
        return ("JackknifeGoodnessFitStat(expected={}, observed={},"
                " bias={}, sd={}, z_score={})").format(
                    self.expected, self.observed,
                    self.bias, self.sd, self.z_score)

    def apply(self, fun):
        return JackknifeGoodnessFitStat(
            fun(self.expected),
            fun(self.observed),
            fun(self.jackknifed_array))

    def __add__(self, other):
        other = self._get_other(other)
        return JackknifeGoodnessFitStat(
            self.expected + other.expected,
            self.observed + other.observed,
            self.jackknifed_array + other.jackknifed_array)

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        return self + (-1)*other

    def __rsub__(self, other):
        return (self * (-1)) + other

    def __mul__(self, other):
        other = self._get_other(other)
        return JackknifeGoodnessFitStat(
            self.expected * other.expected,
            self.observed * other.observed,
            self.jackknifed_array * other.jackknifed_array)

    def __rmul__(self, other):
        return self * other

    def __pow__(self, other):
        other = self._get_other(other)
        return JackknifeGoodnessFitStat(
            self.expected ** other.expected,
            self.observed ** other.observed,
            self.jackknifed_array ** other.jackknifed_array)

    def __rpow__(self, other):
        other = self._get_other(other)
        return other ** self

    def __truediv__(self, other):
        return self * (other**-1)

    def __rtruediv__(self, other):
        return (self**-1) * other

    def _get_other(self, other):
        try:
            other.expected, other.observed, other.jackknifed_array
        except AttributeError:
            return type(self)(other, other, other)
        else:
            return other


def jackknife_arr_op(wrapped_op):
    @ft.wraps(wrapped_op)
    def wraps_op(self, other):
        try:
            other.est, other.jackknife
        except AttributeError:
            return wrapped_op(self, JackknifeStat(other, other))
        else:
            return wrapped_op(self, other)
    return wraps_op


class JackknifeStat(object):
    @classmethod
    def from_chunks(cls, x):
        tot = np.sum(x)
        return cls(tot, tot - x)

    def __init__(self, est, jackknife):
        self.est = est
        self.jackknife = jackknife

    def apply(self, fun):
        return JackknifeStat(fun(self.est),
                             fun(self.jackknife))

    @jackknife_arr_op
    def __add__(self, other):
        return JackknifeStat(self.est + other.est,
                             self.jackknife + other.jackknife)

    def __radd__(self, other):
        return self + other

    def __neg__(self):
        return JackknifeStat(-self.est, -self.jackknife)

    @jackknife_arr_op
    def __sub__(self, other):
        return self + (-other)

    def __rsub__(self, other):
        return -self + other

    @jackknife_arr_op
    def __mul__(self, other):
        return JackknifeStat(self.est * other.est,
                             self.jackknife * other.jackknife)

    def __rmul__(self, other):
        return self * other

    def __truediv__(self, other):
        return self * (other**-1)

    def __rtruediv__(self, other):
        return (self**-1) * other

    @jackknife_arr_op
    def __pow__(self, other):
        return JackknifeStat(self.est ** other.est,
                             self.jackknife ** other.jackknife)

    @jackknife_arr_op
    def __rpow__(self, other):
        return JackknifeStat(other.est ** self.est,
                             other.jackknife ** self.jackknife)

    @property
    def resids(self):
        return self.jackknife - self.est

    @property
    def var(self):
        return np.mean(self.resids**2) * (len(self.jackknife) - 1)

    @property
    def bias(self):
        return np.mean(self.resids) * (len(self.jackknife) - 1)

    @property
    def sd(self):
        return np.sqrt(self.var)

    @property
    def z_score(self):
        return self.est / self.sd

    def __repr__(self):
        return "JackknifeStat(est={}, sd={}) at {}".format(
            self.est, self.sd, hex(id(self)))



import logging
import numpy as np
import pandas as pd
import momi
import json
from collections import Counter
from itertools import product

def sfs_from_dadi(infile, outfile=None):
    """Generate a momi formatted :class:`Sfs` object from the dadi
    frequency spectrum format as documented in section 3.1 of the dadi
    manual. One slight modification is that if you include population
    names after the "folding" flag on the info line, then this
    function will honor these pop names.

    :param str infile: dadi-style sfs to be converted
    :param str,None outfile: Optional output file to write the :class:`Sfs` to

    :rtype: :class:`Sfs`
    """
    dat = open(infile).readlines()
    ## Get rid of comment lines
    dat = [x.strip() for x in dat if not x.startswith("#")]
    if not len(dat) == 3:
        raise Exception("Malformed dadi sfs {}.\n  Must have 3 lines, yours has {}".format(infile, len(dat)))

    ## Parse the info line into nsamps per pop (list of ints), folding flag, and pop names list (if they are given)
    info = dat[0].split()
    nsamps = []
    ## Keep carving off values as long as they cast successfully as int
    for i in info:
        try:
            nsamps.append(int(i))
        except:
            break
    nsamps = np.array(nsamps)
    ## Get rid of quoted pop names if they exist
    pops = [x.replace('"', '') for x in info[len(nsamps)+1:]]
    folded = info[len(nsamps)]
    folded = False if "un" in folded else True
    if not len(pops) == len(nsamps):
        print("Number of populations doesn't agree with number of pop names, using generic names.")
        pops = ["pop"+x for x in range(len(nsamps))]
    logging.getLogger(__name__).info("Info nsamps={} folded={} pops={}".format(nsamps, folded, pops))
    ## Get mask
    mask = list(map(int, dat[2].split()))
    logging.getLogger(__name__).debug("bitmask: {}".format(mask))

    ## Get sfs, and reshape based on sample sizes
    sfs = list(map(float, dat[1].split()))
    logging.getLogger(__name__).debug("SFS from file: {}".format(sfs))
    length = np.ma.array(sfs, mask=mask).sum()
    sfs = np.array(sfs).reshape(nsamps)
    logging.getLogger(__name__).info("length {}".format(length))
    logging.getLogger(__name__).debug("Reshaped sfs: {}".format(sfs))

    ## Get counts per sfs bin
    counts = Counter()
    for sfsbin in product(*[range(y) for y in [x for x in nsamps]]):
        ## Ignore monomorphic snps
        ## nsamps minus 1 here because of the off by one diff between number
        ## of bins in the sfs and actual number of samples
        if sfsbin == tuple(nsamps-1) or sfsbin == tuple([0] * len(nsamps)):
            continue
        ## ignore zero bin counts
        if sfs[sfsbin] == 0:
            continue
        logging.getLogger(__name__).debug(sfsbin, sfs[sfsbin]),
        counts.update({sfsbin:sfs[sfsbin]})
    logging.getLogger(__name__).debug("nbins {}".format(len(counts)))

    ## Convert SFS bin style into momi config style
    configs = pd.DataFrame(index=range(len(counts)), columns=pops)
    
    locus_info = []
    for i, c in enumerate(counts):
        ## (n-1) here because nbins in dadi sfs is n+1
        cfg = np.array([[(n-1)-x, x] for x,n in zip(c, nsamps)])
        configs.iloc[i] = [list(map(int, list(x))) for x in cfg]
        locus_info.append([0, i, counts[c]])
    logging.getLogger(__name__).info("n_snps {}".format(np.sum([x[2] for x in locus_info])))

    ## Finally build the momi style sfs dictionary and write it out
    momi_sfs = {"sampled_pops":pops,
        "folded":folded,
        "length":int(length),
        "configs":configs.values.tolist(),
        "(locus,config_id,count)":locus_info}

    if outfile == None:
        outfile = "./.tmp.sfs"
    with open(outfile, 'w') as out:
        out.write("{}".format(json.dumps(momi_sfs)))
    ## make it pretty
    sfs = momi.Sfs.load(outfile)
    ## Fold if unfolded
    if folded: sfs = sfs.fold()
    sfs.dump(outfile)

    return sfs



import momi.size_history as size_history
import momi.util as util
import numpy as np
from scipy.special import comb as binom
import random


def q(n, b):
    return sum([k / binom(k, 2) * binom(n - b - 1, k - 2) / binom(n - 1, k - 1) / 2.0 for k in range(2, n + 1)])


def test_constant_sfs():
    n_max = 10
    hist = size_history.ConstantHistory(float("inf"), 1.0)
    for b in range(1, n_max):
        assert abs(hist.sfs(n_max)[b] - q(n_max, b)) < 1e-8




#!/usr/bin/env python

from setuptools import setup, Extension
import os
import sys

extensions = []
setup_requires = []
install_requires = ['cached_property>=1.3']

use_cython = "USE_CYTHON" in os.environ
if use_cython:
    from Cython.Build import cythonize
    ext = ".pyx"
else:
    ext = ".c"

class numpy_get_include(object):
    def __str__(self):
        import numpy
        return numpy.get_include()
numpy_get_include = numpy_get_include()

on_rtd = os.environ.get('READTHEDOCS') == 'True'
if not on_rtd:
    extra_compile_args=["-fopenmp"]
    extra_link_args=["-fopenmp"]

    if use_cython:
        numpy_get_include = str(numpy_get_include)

    extensions = [
        Extension("momi.convolution",
                  sources=["momi/convolution" + ext],
                  extra_compile_args=extra_compile_args,
                  extra_link_args=extra_link_args,
                  include_dirs=[numpy_get_include]),
        Extension("momi.einsum2.parallel_matmul",
                  sources=["momi/einsum2/parallel_matmul" + ext],
                  extra_compile_args=extra_compile_args,
                  extra_link_args=extra_link_args,
                  include_dirs=[numpy_get_include]),
        Extension("momi.w_matrix",
                  sources=["momi/w_matrix" + ext],
                  extra_compile_args=extra_compile_args,
                  extra_link_args=extra_link_args,
                  include_dirs=[numpy_get_include])
        ]

    numpy_minver = '1.9.0'

    if use_cython:
        extensions = cythonize(extensions)
        setup_requires = ['cython']
    else:
        setup_requires = [f'numpy>={numpy_minver}']

    install_requires.extend([
        'autograd>=1.2.0', f'numpy>={numpy_minver}', 'networkx', 'scipy',
        'pandas', 'msprime', "matplotlib>=3.3", "seaborn", "pysam"])

setup(name='momi',
      version='2.1.21',
      description='MOran Model for Inference',
      author='Jack Kamm, Jonathan Terhorst, Richard Durbin, Yun S. Song',
      author_email='jkamm@stat.berkeley.edu, terhorst@stat.berkeley.edu, yss@eecs.berkeley.edu',
      packages=['momi', 'momi.einsum2', 'momi.data'],
      setup_requires=setup_requires,
      install_requires=install_requires,
      python_requires='>=3.5',
      keywords=['population genetics', 'statistics',
                'site frequency spectrum', 'coalescent'],
      url='https://github.com/jackkamm/momi2',
      ext_modules=extensions)



import autograd
import autograd.numpy as np
from autograd.extend import primitive, defvjp
from .parallel_matmul import _par_matmul

@primitive
def batched_dot(a, b):
    if len(a.shape) != 3 or len(b.shape) != 3 or a.shape[0] != b.shape[0]:
        raise ValueError("a,b must be 3-dimensional arrays, with a.shape[0]==b.shape[0] and a.shape[2]==b.shape[1]")
    elif a.shape[0] == 1:
        ## use numpy.dot for blas
        a = np.reshape(a, a.shape[1:])
        b = np.reshape(b, b.shape[1:])
        c = np.dot(a, b)
        return np.reshape(c, [1] + list(c.shape))
    elif a.shape[2] == 1:
        ## the main cost is simply allocating space for the array,
        ## so we are better off doing things in serial
        a = np.reshape(a, a.shape[:-1])
        b = np.reshape(b, (b.shape[0], b.shape[2]))
        if a.shape[-1] > 1 and b.shape[-1] > 1:
            ## batch outer product
            return np.einsum("ij,ik->ijk", a, b)
        else:
            ## broadcasted elementary-wise multiplication
            outshape = (a.shape[0], a.shape[1], b.shape[1])
            a = np.transpose(a)
            b = np.transpose(b)
            if a.shape[0] == 1:
                a = np.reshape(a, [-1])
            if b.shape[0] == 1:
                b = np.reshape(b, [-1])
            return np.transpose(np.reshape(a*b, outshape[::-1]))
    else:
        ## parallel batched matrix multiply
        return _par_matmul(a, b)

defvjp(
    batched_dot,
    lambda ans, a, b: lambda g: batched_dot(g,
                                            np.transpose(b, (0,2,1))),
    lambda ans, a, b: lambda g: batched_dot(np.transpose(a, (0,2,1)),
                                            g))
#batched_dot.defgrad(lambda ans, a, b: lambda g: batched_dot(g, np.transpose(b, (0,2,1))))
#batched_dot.defgrad(lambda ans, a, b: lambda g: batched_dot(np.transpose(a, (0,2,1)), g), argnum=1)
#batched_dot.defvjp(lambda g, ans, vs, gvs, a, b: batched_dot(
#    g, np.transpose(b, (0,2,1))))
#batched_dot.defvjp(
#    lambda g, ans, vs, gvs, a, b: batched_dot(
#        np.transpose(a, (0,2,1)), g),
#    argnum=1)

def einsum2(*args, **kwargs):
    """
    einsum2(subscripts_str, arr0, arr1)
    or,
    einsum2(op0, subscript_list0, arr1, subscript_list1,
            output_subscript_list)

    This function is similar to einsum, except it only operates
    on two input arrays, does not allow diagonal operations
    (repeated subscripts on the same array), and requires the output
    subscripts to always be specified. Also, when specifying
    subscripts via lists, the subscripts can be arbitrary keys
    (unlike numpy.einsum, where they have to be integers).

    Unlike the standard einsum, einsum2 will perform computations
    in parallel. The number of parallel threads is selected automatically,
    but you can also control this with the environment variable
    OMP_NUM_THREADS.

    To perform the parallel computation, einsum2 will either use
    numpy.dot (if possible), otherwise it will use a parallel
    for loop. The advantage of using numpy.dot is that it
    uses BLAS which is much faster than a for loop. However,
    you need to make sure numpy is compiled against a parallel BLAS
    implementation such as MKL or OpenBlas. You won't need to worry
    about this for most packaged, precompiled versions of numpy
    (e.g. Anaconda Python).
    """
    if isinstance(args[0], str):
        subscripts, a, b = args[:3]
        ab_subs, out_subs = subscripts.split("->")
        a_subs, b_subs = ab_subs.split(",")
        return _einsum2(a, list(a_subs), b, list(b_subs), list(out_subs), *args[3:], **kwargs)
    else:
        return _einsum2(*args, **kwargs)

def _einsum2(a, a_sublist, b, b_sublist, out_sublist):
    for subs in a_sublist, b_sublist, out_sublist:
        if len(subs) != len(set(subs)):
            raise NotImplementedError("Repeated subscripts not implemented")

    a, a_sublist = _sum_unique_axes(a, a_sublist, b_sublist, out_sublist)
    b, b_sublist = _sum_unique_axes(b, b_sublist, a_sublist, out_sublist)

    a_subs, b_subs, out_subs = map(set, (a_sublist, b_sublist, out_sublist))
    if out_subs - (a_subs | b_subs):
        raise ValueError("Output subscripts must be contained within input subscripts")

    a_minus_b = list(a_subs - b_subs)
    b_minus_a = list(b_subs - a_subs)
    # _sum_unique_axes should have removed any axes unique to a,b
    assert set(a_minus_b) <= out_subs and set(b_minus_a) <= out_subs

    ab = a_subs & b_subs
    abc = list(ab & out_subs)
    ab_minus_c = list(ab - out_subs)

    shapes = {}
    for arr,sublist in ((a,a_sublist), (b,b_sublist)):
        # arr.shape breaks in autograd if it has no dimension
        if sublist:
            for i,s in zip(arr.shape, sublist):
                if s not in shapes:
                    shapes[s] = i
                elif shapes[s] != i:
                    raise ValueError("a,b shapes don't match")

    a = _reshape(a, a_sublist, abc, a_minus_b, ab_minus_c)
    b = _reshape(b, b_sublist, abc, ab_minus_c, b_minus_a)
    assert len(a.shape) == len(b.shape) == 3

    c = batched_dot(a, b)

    c_sublist = abc + a_minus_b + b_minus_a
    c = np.reshape(c, [shapes[s] for s in c_sublist])

    return _transpose(c, c_sublist, out_sublist)

def einsum1(in_arr, in_sublist, out_sublist):
    in_arr, in_sublist = _sum_unique_axes(in_arr, in_sublist, out_sublist)
    return _transpose(in_arr, in_sublist, out_sublist)

def _reshape(in_arr, in_sublist, *out_sublists):
    assert len(out_sublists) == 3

    old_sublist = in_sublist
    in_sublist = sum(out_sublists, [])
    in_arr = _transpose(in_arr, old_sublist, in_sublist)

    # in_arr.shape breaks in autograd if it has no dimension
    if in_sublist:
        shapes = {s:i for i,s in zip(in_arr.shape, in_sublist)}
    else: shapes = {}
    return np.reshape(in_arr, [np.prod([shapes[s] for s in out_subs], dtype=int)
                               for out_subs in out_sublists])

def _transpose(in_arr, in_sublist, out_sublist):
    if set(in_sublist) != set(out_sublist):
        raise ValueError("Input and output subscripts don't match")
    for sublist in (in_sublist, out_sublist):
        if len(set(sublist)) != len(sublist):
            raise NotImplementedError("Repeated subscripts not implemented")
    in_idxs = {k:v for v,k in enumerate(in_sublist)}
    return np.transpose(in_arr, axes=[in_idxs[s] for s in out_sublist])

def _sum_unique_axes(in_arr, in_sublist, *keep_subs):
    # assume no repeated subscripts
    assert len(in_sublist) == len(set(in_sublist))

    out_sublist = []
    sum_axes = []
    keep_subs = set([s for ks in keep_subs for s in ks])
    for idx, sub in enumerate(in_sublist):
        if sub in keep_subs:
            out_sublist.append(sub)
        else:
            sum_axes.append(idx)
    if sum_axes:
        return np.sum(in_arr, axis=tuple(sum_axes)), out_sublist
    else:
        return in_arr, out_sublist



import os
import itertools as it
from collections import defaultdict, Counter
import json
import re
import gzip
import logging
import numpy as np
import pysam
from cached_property import cached_property
from .configurations import ConfigList
from .sfs import Sfs
from ..util import memoize_instance
from .compressed_counts import (
    CompressedAlleleCounts, _CompressedHashedCounts, _CompressedList)


logger = logging.getLogger(__name__)


def snp_allele_counts(chrom_ids, positions, populations,
                      ancestral_counts, derived_counts,
                      length=None, use_folded_sfs=False):
    """Create a :class:`SnpAlleleCounts` object.

    :param iterator chrom_ids: the CHROM at each SNP
    :param iterator positions: the POS at each SNP
    :param list populations: the population names

    :param iterator ancestral_counts: iterator over tuples of \
    the ancestral counts at each SNP. tuple length should \
    be the same as the number of populations

    :param iterator derived_counts: iterator over tuples of \
    the derived counts at each SNP. tuple length should \
    be the same as the number of populations

    :param bool use_folded_sfs: whether the folded SFS should \
    be used when computing likelihoods. Set this to True if there is \
    uncertainty about the ancestral allele.

    :rtype: :class:`SnpAlleleCounts`
    """
    config_iter = (tuple(zip(a, d)) for a, d in
                   zip(ancestral_counts, derived_counts))
    chrom_ids = _CompressedList(chrom_ids)
    positions = list(positions)
    return SnpAlleleCounts(chrom_ids, positions,
                           CompressedAlleleCounts.from_iter(
                               config_iter, len(populations)),
                           populations, use_folded_sfs,
                           [], length,
                           len(chrom_ids), 0)


class SnpAlleleCounts(object):
    """
    The allele counts for a list of SNPs.

    To create a :class:`SnpAlleleCounts` object,
    use :func:`SnpAlleleCounts.read_vcf`, :func:`SnpAlleleCounts.load`,
    or :func:`snp_allele_counts`. Do NOT call the class constructor directly,
    it is for internal use only.
    """
    @classmethod
    def read_vcf(cls, vcf_file, ind2pop,
                 bed_file=None, ancestral_alleles=True,
                 info_aa_field="AA"):
        """Read in a VCF file and return the allele counts at biallelic SNPs.

        :param str vcf_file: VCF file to read in. The VCF file should be indexed, \
        e.g. with tabix or bcftools. Alternatively, if equal to "-", reads \
        the VCF from stdin.
        :param dict ind2pop: Maps individual samples to populations.
        :param str,None bed_file: BED accessibility regions file. \
        Only regions in the BED file are read from the VCF. The BED file \
        is also used to determine the size of the data in bases, so the same \
        BED file should NOT be used across multiple VCFs (otherwise regions will be double \
        counted towards the data length). If no BED is provided, all SNPs in the VCF are read, and the length \
        of the data is set to be unknown.
        :param bool,str ancestral_alleles: If True, use the AA INFO field to \
        determine ancestral allele, skipping SNPs missing this field. \
        If False, ignore ancestral allele information, and set the \
        ``SnpAlleleCounts.use_folded_sfs`` property so that \
        the folded SFS is used by default when computing likelihoods. \
        Finally, if ``ancestral_alleles`` is a string that is the name \
        of a population in ``ind2pop``, then treat that population as an outgroup, \
        using its consensus allele as the ancestral allele; SNPs without \
        consensus are skipped.
        :param str info_aa_field: The INFO field to read Ancestral Allele from. \
        Default is "AA". Only has effect if ``ancestral_alleles=True``.

        :rtype: :class:`SnpAlleleCounts`
        """
        bcf_in = pysam.VariantFile(vcf_file)

        # subset samples for faster VCF parsing
        bcf_in.subset_samples(list(ind2pop.keys()))
        samples = list(bcf_in.header.samples)
        assert set(samples) == set(ind2pop.keys())

        # extract populations, samples
        pop2idxs = defaultdict(list)
        for ind, pop in ind2pop.items():
            pop2idxs[pop].append(samples.index(ind))
        sampled_pops = sorted(
            p for p in pop2idxs.keys() if p != ancestral_alleles)

        # objects to store chrom, pos, configs
        chrom_list = _CompressedList()
        pos_list = []
        compressed_hashed = _CompressedHashedCounts(len(sampled_pops))
        excluded = []

        # pre-allocate objects for temporary configs
        pop_allele_counts = {pop: Counter() for pop in pop2idxs.keys()}
        config = np.zeros((len(sampled_pops), 2), dtype=int)

        if bed_file:
            def open_bed():
                if bed_file.endswith(".gz"):
                    return gzip.open(bed_file, "rt")
                else:
                    return open(bed_file)
            length = 0
            with open_bed() as bed:
                for line in bed:
                    line = line.split()
                    contig = line[0]
                    start, end = map(int, line[1:3])
                    length += (end - start)
                    fetcher = bcf_in.fetch(
                        contig, start, end)
                    cls._read_vcf_helper(
                        fetcher, chrom_list, pos_list,
                        compressed_hashed, excluded,
                        ancestral_alleles, pop2idxs,
                        sampled_pops, pop_allele_counts, config,
                        info_aa_field)
        else:
            length = None
            logger.warn("No BED provided, will need to specify length"
                        " manually with mutation rate")
            fetcher = bcf_in.fetch()
            cls._read_vcf_helper(
                fetcher, chrom_list, pos_list, compressed_hashed,
                excluded, ancestral_alleles, pop2idxs, sampled_pops,
                pop_allele_counts, config, info_aa_field)

        if len(compressed_hashed) == 0:
            logger.warn("No valid SNPs read! Try setting "
                        "ancestral_alleles=False.")

        return cls(chrom_list, pos_list,
                   compressed_hashed.compressed_allele_counts(),
                   sampled_pops, not ancestral_alleles, [], length,
                   len(chrom_list), len(excluded))

    @classmethod
    def _read_vcf_helper(
            cls, bcf_in_fetch, chrom, pos, compressed_hashed, excluded,
            ancestral_alleles, pop2idxs, sampled_pops,
            pop_allele_counts, config, info_aa_field):
        for rec in bcf_in_fetch:
            if len(rec.alleles) != 2:
                continue

            for pop, inds in pop2idxs.items():
                pop_allele_counts[pop].clear()
                for i in inds:
                    for a in rec.samples[i].allele_indices:
                        if a is not None:
                            pop_allele_counts[pop][a] += 1

            if ancestral_alleles is True:
                try:
                    aa = rec.info[info_aa_field]
                except KeyError:
                    excluded.append((rec.chrom, rec.pos))
                    continue
                else:
                    try:
                        aa = rec.alleles.index(aa)
                    except ValueError:
                        excluded.append((rec.chrom, rec.pos))
                        continue
            elif ancestral_alleles:
                outgroup_counts = pop_allele_counts[ancestral_alleles]
                if len(outgroup_counts) != 1:
                    excluded.append((rec.chrom, rec.pos))
                    continue
                aa, = outgroup_counts.keys()
            else:
                aa = 0

            config *= 0
            for pop_idx, pop in enumerate(sampled_pops):
                for a, n in pop_allele_counts[pop].items():
                    config[pop_idx, a] = n

            if aa == 1:
                config = config[:, ::-1]

            compressed_hashed.append(config)

            chrom.append(rec.chrom)
            pos.append(rec.pos)

            if len(pos) % 10000 == 0:
                logger.info("Read vcf up to CHR {}, POS {}".format(
                    chrom[-1], pos[-1]))

    @classmethod
    def concatenate(cls, to_concatenate):
        """Combine a list of :class:`SnpAlleleCounts` into a single object.

        :param iterator to_concatenate: Iterator over :class:`SnpAlleleCounts`
        :rtype: :class:`SnpAlleleCounts`
        """
        to_concatenate = iter(to_concatenate)
        first = next(to_concatenate)
        populations = list(first.populations)
        nonascertained = list(first.non_ascertained_pops)
        to_concatenate = it.chain([first], to_concatenate)

        #chrom_ids = []
        chrom_ids = _CompressedList()
        positions = []
        index2uniq = []

        compressed_hashes = _CompressedHashedCounts(len(populations))

        use_folded_sfs = False
        length = 0
        n_read_snps = 0
        n_excluded_snps = 0
        for snp_cnts in to_concatenate:
            use_folded_sfs = (use_folded_sfs or
                                     snp_cnts.use_folded_sfs)

            if any([list(snp_cnts.populations) != populations,
                    list(snp_cnts.non_ascertained_pops) != nonascertained]):
                raise ValueError(
                    "Datasets must have same populations with same"
                    " ascertainment to concatenate")
            old2new_uniq = []
            for config in snp_cnts.compressed_counts.config_array:
                compressed_hashes.append(config)
                old2new_uniq.append(compressed_hashes.index2uniq(-1))

            assert len(snp_cnts.chrom_ids) == len(snp_cnts.compressed_counts.index2uniq)
            assert len(snp_cnts.chrom_ids) == len(snp_cnts.positions)
            for chrom, pos, old_uniq in zip(snp_cnts.chrom_ids,
                                            snp_cnts.positions,
                                            snp_cnts.compressed_counts.index2uniq):
                chrom_ids.append(chrom)
                positions.append(pos)
                index2uniq.append(old2new_uniq[old_uniq])

            try:
                length += snp_cnts.length
            except TypeError:
                length = None

            n_read_snps += snp_cnts.n_read_snps
            n_excluded_snps += snp_cnts.n_excluded_snps

            for k, v in Counter(snp_cnts.chrom_ids).items():
                logger.info("Added {} SNPs from Chromosome {}".format(v, k))

        # make sure the positions are sorted
        chrom_ids, positions, index2uniq = zip(*sorted(zip(
            chrom_ids, positions, index2uniq)))

        compressed_counts = CompressedAlleleCounts(
            compressed_hashes.config_array(), index2uniq)
        ret = cls(chrom_ids, positions, compressed_counts, populations,
                  use_folded_sfs=use_folded_sfs,
                  non_ascertained_pops=nonascertained,
                  length=length,
                  n_read_snps=n_read_snps,
                  n_excluded_snps=n_excluded_snps)
        logger.info("Finished concatenating datasets")
        return ret

    @classmethod
    def load(cls, f):
        """Load :class:`SnpAlleleCounts` created \
        from :meth:`SnpAlleleCounts.dump` or ``python -m momi.read_vcf ...``

        :param str,file f: file object or file name to read in
        :rtype: :class:`SnpAlleleCounts`
        """
        if isinstance(f, str):
            if f.endswith(".gz"):
                with gzip.open(f, "rt") as gf:
                    return cls.load(gf)
            else:
                with open(f) as gf:
                    return cls.load(gf)

        # default values for back-compatibility
        items = {"use_folded_sfs": False,
                 "non_ascertained_pops": [],
                 "length": None,
                 "n_excluded_snps": 0}
        items_re = re.compile(r'\s*"(.*)":\s*(.*),\s*\n')
        config_re = re.compile(r'\s*"configs":\s*\[\s*\n')
        chrom_pos_idx_re = re.compile(
            r'(\s*)"\(chrom_id,position,config_id\)":(\s*)\[(\s*)\n')
        for line in f:
            items_matched = items_re.match(line)
            config_matched = config_re.match(line)
            chrom_pos_idx_matched = chrom_pos_idx_re.match(line)
            if chrom_pos_idx_matched:
                logger.info("Reading SNPs")
                line_re = re.compile(r'\s*\[(.*)\],?\s*\n')
                #chrom_ids = []
                chrom_ids = _CompressedList()
                positions = []
                config_ids = []
                for i, line in enumerate(f):
                    if i % 100000 == 0 and i > 0:
                        logger.info(i)
                    try:
                        curr = line_re.match(line).group(1)
                    except AttributeError:
                        assert line == "\t]\n"
                        break
                    else:
                        chrom, pos, idx = curr.split(",")
                        chrom = chrom.strip()
                        assert chrom[0] == chrom[-1] == '"'
                        chrom_ids.append(chrom[1:-1])
                        positions.append(float(pos))
                        config_ids.append(int(idx))
                logger.info("Finished reading SNPs")
            elif config_matched:
                line_re = re.compile(r'\s*\[(.*)\],?\s*\n')
                configs = []
                logger.info("Reading unique configs")
                for i, line in enumerate(f):
                    if i % 100000 == 0:
                        if i > 0:
                            logger.info(i)
                            configs[-1] = np.array(configs[-1],
                                                   dtype=int)
                        configs.append([])
                    try:
                        conf = line_re.match(line).group(1)
                    except AttributeError:
                        assert line == "\t],\n"
                        break
                    else:
                        #configs[-1].append(ast.literal_eval(conf))
                        # ast.literal_eval is slow
                        conf = conf.replace("[", " ")
                        conf = conf.replace(",", " ")
                        conf = [[int(x_i) for x_i in x.split()]
                                for x in conf.split("]")[:-1]]
                        configs[-1].append(conf)
                configs[-1] = np.array(configs[-1], dtype=int)
                configs = np.concatenate(configs)
                logger.info("Finished reading configs")
            elif items_matched:
                items[items_matched.group(1)] = json.loads(
                    items_matched.group(2))

        # for back-compatibility
        if "n_read_snps" not in items:
            items["n_read_snps"] = len(chrom_ids)

        logging.debug("Creating CompressedAlleleCounts")
        compressed_counts = CompressedAlleleCounts(
            configs, config_ids,
            sort=False)

        logging.debug("Creating SnpAlleleCounts")
        return cls(chrom_ids, positions, compressed_counts,
                   **items)

    def dump(self, f):
        """Write data in JSON format.

        :param str,file f: filename or file object. \
        If a filename, the resulting file is gzipped.
        """
        if isinstance(f, str):
            if f.endswith(".gz"):
                with gzip.open(f, "wt") as gf:
                    self.dump(gf)
            else:
                with open(f, "w") as gf:
                    self.dump(gf)
            return

        print("{", file=f)
        print('\t"populations": {},'.format(
            json.dumps(list(self.populations))), file=f)
        if self.non_ascertained_pops:
            print('\t"non_ascertained_pops": {},'.format(
                json.dumps(list(self.non_ascertained_pops))), file=f)
        print('\t"use_folded_sfs": {},'.format(
            json.dumps(self.use_folded_sfs)), file=f)
        print('\t"length": {},'.format(
            json.dumps(self.length)), file=f)
        print('\t"n_read_snps": {},'.format(self.n_read_snps), file=f)
        print('\t"n_excluded_snps": {},'.format(self.n_excluded_snps), file=f)
        print('\t"configs": [', file=f)
        n_configs = len(self.compressed_counts.config_array)
        for i, c in enumerate(self.compressed_counts.config_array.tolist()):
            if i != n_configs - 1:
                print("\t\t{},".format(c), file=f)
            else:
                # no trailing comma
                print("\t\t{}".format(c), file=f)
        print("\t],", file=f)
        print('\t"(chrom_id,position,config_id)": [', file=f)
        n_positions = len(self)
        for i, chrom_id, pos, config_id in zip(
                range(n_positions), list(self.chrom_ids),
                self.positions.tolist(),
                self.compressed_counts.index2uniq.tolist()):
            if i != n_positions - 1:
                print("\t\t{},".format(json.dumps(
                    (chrom_id, pos, config_id))), file=f)
            else:
                # no trailling comma
                print("\t\t{}".format(json.dumps(
                    (chrom_id, pos, config_id))), file=f)
        print("\t]", file=f)
        print("}", file=f)

    def __init__(self, chrom_ids, positions,
                 compressed_counts, populations,
                 use_folded_sfs, non_ascertained_pops, length,
                 n_read_snps, n_excluded_snps):
        if any([len(compressed_counts) != len(chrom_ids),
                len(chrom_ids) != len(positions)]):
            raise IOError(
                "chrom_ids, positions, allele_counts should have same length")

        self.chrom_ids = chrom_ids
        self.positions = np.array(positions)
        self.compressed_counts = compressed_counts
        self.populations = populations
        self.non_ascertained_pops = non_ascertained_pops
        self.use_folded_sfs = use_folded_sfs
        self.length = length
        self.n_read_snps = n_read_snps
        self.n_excluded_snps = n_excluded_snps

    def __eq__(self, other):
        try:
            return (
                self.compressed_counts == other.compressed_counts and
                np.all(self.chrom_ids == other.chrom_ids) and
                np.all(self.positions == other.positions) and
                self.length == other.length and
                (self.use_folded_sfs ==
                 other.use_folded_sfs) and
                self.n_read_snps == other.n_read_snps and
                self.n_excluded_snps == other.n_excluded_snps)
        except AttributeError:
            return False

    @memoize_instance
    def _chunk_data(self, n_chunks):
        chunk_len = len(self.chrom_ids) / float(n_chunks)
        new_pos = list(range(len(self.chrom_ids)))
        new_chrom = _CompressedList(
            int(np.floor(i / chunk_len)) for i in new_pos)
        return SnpAlleleCounts(
            new_chrom, new_pos, self.compressed_counts,
            self.populations, self.use_folded_sfs,
            self.non_ascertained_pops, self.length,
            self.n_read_snps, self.n_excluded_snps)

    def extract_sfs(self, n_blocks):
        """Extracts SFS from data.

        :param int n_blocks: Number of blocks to split SFS into, for jackknifing and bootstrapping
        :rtype: :class:`Sfs`
        """
        if n_blocks is None:
            return self._sfs
        else:
            return self._chunk_data(n_blocks)._sfs

    @property
    def p_missing(self):
        return self._sfs.p_missing

    def subset_populations(self, populations, non_ascertained_pops=None):
        if non_ascertained_pops is not None:
            non_ascertained_pops = tuple(non_ascertained_pops)
        return self._subset_populations(tuple(populations),
                                        non_ascertained_pops)

    @memoize_instance
    def _subset_populations(self, populations, non_ascertained_pops):
        if non_ascertained_pops is None:
            non_ascertained_pops = [p for p in self.non_ascertained_pops
                                    if p in populations]

        newPopIdx_to_oldPopIdx = np.array([
            self.populations.index(p) for p in populations], dtype=int)

        uniq_new_configs = CompressedAlleleCounts.from_iter(
            self.compressed_counts.config_array[:, newPopIdx_to_oldPopIdx, :],
            len(populations), sort=False)

        new_compressed_configs = CompressedAlleleCounts(
            uniq_new_configs.config_array,
            [uniq_new_configs.index2uniq[i]
             for i in self.compressed_counts.index2uniq],
            sort=False)

        return SnpAlleleCounts(
            self.chrom_ids, self.positions,
            new_compressed_configs, populations,
            self.use_folded_sfs,
            non_ascertained_pops, self.length,
            self.n_read_snps, self.n_excluded_snps)

    def __getitem__(self, i):
        return self.compressed_counts[i]

    def __len__(self):
        return len(self.compressed_counts)

    def filter(self, idxs):
        return SnpAlleleCounts(self.chrom_ids[idxs],
                               self.positions[idxs],
                               self.compressed_counts.filter(idxs),
                               self.populations,
                               self.use_folded_sfs,
                               self.non_ascertained_pops, self.length,
                               self.n_read_snps, self.n_excluded_snps)

    def down_sample(self, sampled_n_dict):
        """Randomly subsample the alleles at each site.

        :param dict sampled_n_dict: dictionary mapping population names to the subsample size (number of alleles to subsample from each population).
        :rtype: :class:`SnpAlleleCounts`
        """
        pops, sub_n = zip(*sampled_n_dict.items())
        pop_idxs = [self.populations.index(p) for p in pops]

        def sub_counts():
            for config in self:
                config = list(config)
                for i, n in zip(pop_idxs, sub_n):
                    curr_n = sum(config[i])
                    if curr_n > n:
                        old_anc, old_der = config[i]
                        new_anc = np.random.hypergeometric(
                            old_anc, old_der, n)
                        new_der = n - new_anc
                        config[i] = (new_anc, new_der)
                yield config

        return SnpAlleleCounts(
            self.chrom_ids, self.positions,
            CompressedAlleleCounts.from_iter(
                sub_counts(), len(self.populations)),
            self.populations, self.use_folded_sfs,
            self.non_ascertained_pops, self.length,
            self.n_read_snps, self.n_excluded_snps)

    @property
    def is_polymorphic(self):
        configs = self.compressed_counts.config_array
        ascertained_only = configs[:, self.ascertainment_pop, :]
        ascertained_is_poly = (ascertained_only.sum(axis=1) != 0).all(axis=1)
        return ascertained_is_poly[self.compressed_counts.index2uniq]

    @property
    def ascertainment_pop(self):
        return np.array([(pop not in self.non_ascertained_pops)
                         for pop in self.populations])

    #@property
    #def seg_sites(self):
    #    try:
    #        self._seg_sites
    #    except:
    #        filtered = self.filter(self.is_polymorphic)
    #        idx_list = [
    #            np.array([i for chrom, i in grouped_idxs])
    #            for key, grouped_idxs in it.groupby(
    #                    zip(filtered.chrom_ids,
    #                        filtered.compressed_counts.index2uniq),
    #                    key=lambda x: x[0])]
    #        self._seg_sites = SegSites(ConfigList(
    #            self.populations,
    #            filtered.compressed_counts.config_array,
    #            ascertainment_pop=self.ascertainment_pop
    #        ), idx_list)
    #    return self._seg_sites

    @cached_property
    def _sfs(self):
        filtered = self.filter(self.is_polymorphic)
        idx_list = [
            np.array([i for chrom, i in grouped_idxs])
            for key, grouped_idxs in it.groupby(
                    zip(filtered.chrom_ids,
                        filtered.compressed_counts.index2uniq),
                    key=lambda x: x[0])]
        configs = ConfigList(
            self.populations,
            filtered.compressed_counts.config_array,
            ascertainment_pop=self.ascertainment_pop)
        if self.length:
            length = self.length * (1 - self._p_excluded)
        else:
            length = None
        ret = Sfs(idx_list, configs, folded=False, length=length)
        if self.use_folded_sfs:
            ret = ret.fold()
        return ret

    @property
    def configs(self):
        return self._sfs.configs

    @property
    def _p_excluded(self):
        n_read = self.n_read_snps
        n_exclude = self.n_excluded_snps

        return n_exclude / (n_read + n_exclude)




import pytest
import momi
import momi.likelihood
from momi import SfsLikelihoodSurface
from demo_utils import simple_five_pop_demo

import autograd.numpy as np
from autograd import grad, hessian, hessian_vector_product, jacobian
import autograd

def test_batches():
    demo = simple_five_pop_demo()

    sampled_n_dict = dict(zip(demo.leafs, [10]*5))
    num_bases=1000
    sfs = demo.simulate_data(
        length=num_bases,
        muts_per_gen=.1/num_bases,
        recoms_per_gen=0,
        num_replicates=1000,
        sampled_n_dict=sampled_n_dict)._sfs


    sfs_len = sfs.n_nonzero_entries

    print("total entries", sfs_len)
    print("total snps", sfs.n_snps())

    assert sfs_len > 30

    demo = demo._get_demo(sampled_n_dict)
    assert np.isclose(SfsLikelihoodSurface(sfs, batch_size=5).log_lik(demo),
                      momi.likelihood._composite_log_likelihood(sfs, demo))

    assert np.allclose(SfsLikelihoodSurface(sfs, batch_size=5).log_lik(demo, vector=True),
                      momi.likelihood._composite_log_likelihood(sfs, demo, vector=True))



# TODO does this test still make sense?
# uses obsolete style of demo functions which I think makes it slow
def test_batches_grad():
    x0 = np.random.normal(size=30)
    pre_demo_func = lambda *x: simple_five_pop_demo(
        x=np.array(x))
    demo = pre_demo_func(*x0)
    pops = demo.leafs

    sampled_n_dict = dict(zip(demo.leafs, [10]*5))
    num_bases=1000
    sfs = demo.simulate_data(
        length=num_bases,
        muts_per_gen=.1/num_bases,
        recoms_per_gen=0,
        num_replicates=1000,
        sampled_n_dict=sampled_n_dict)._sfs


    sfs_len = sfs.n_nonzero_entries

    print("total entries", sfs_len)
    print("total snps", sfs.n_snps())

    assert sfs_len > 30

    sampled_n_dict = dict(zip(pops, [10]*5))
    demo_func = lambda *x: pre_demo_func(*x)._get_demo(sampled_n_dict)

    jac1 = -sfs.n_snps() * grad(SfsLikelihoodSurface(sfs, batch_size=5, demo_func=demo_func, mut_rate=1.).kl_div)(x0)
    jac2 = grad(lambda x: momi.likelihood._composite_log_likelihood(sfs, demo_func(*x), mut_rate=1.))(x0)
    assert np.allclose(jac1, jac2)

# TODO reenable these tests?
#def test_batches_jac():
#    x0 = np.random.normal(size=30)
#    pre_demo_func = lambda *x: simple_five_pop_demo(
#        x=np.array(x), n_lins=(10, 10, 10, 10, 10))
#    demo_func = lambda *x: pre_demo_func(*x).demo_hist.rescaled()
#    pre_demo = pre_demo_func(*x0)
#
#    mu = 10
#    num_bases = 1000
#    sfs = pre_demo.demo_hist.simulate_data(
#        pre_demo.pops, pre_demo.n,
#        mutation_rate=mu/num_bases,
#        recombination_rate=0,
#        length=num_bases,
#        num_replicates=10).sfs
#
#
#    sfs_len = sfs.n_nonzero_entries
#
#    print("total entries", sfs_len)
#    print("total snps", sfs.n_snps())
#
#    assert sfs_len > 30
#
#    jac1 = jacobian(lambda x: SfsLikelihoodSurface(sfs, batch_size=5, demo_func=demo_func, mut_rate=mu).log_lik(x, vector=True))(x0)
#    jac2 = jacobian(lambda x: momi.likelihood._composite_log_likelihood(sfs, demo_func(*x), mut_rate=mu, vector=True))(x0)
#    assert np.allclose(jac1, jac2)
#    print(jac1)
#
#
#
#def test_batches_hess():
#    x0 = np.random.normal(size=30)
#    pre_demo_func = lambda *x: simple_five_pop_demo(
#        x=np.array(x), n_lins=(10, 10, 10, 10, 10))
#    demo_func = lambda *x: pre_demo_func(*x).demo_hist.rescaled()
#    pre_demo = pre_demo_func(*x0)
#
#    mu = .05
#    num_bases = 1000
#    sfs = pre_demo.demo_hist.simulate_data(
#        pre_demo.pops, pre_demo.n,
#        mutation_rate=mu/num_bases,
#        recombination_rate=0,
#        length=num_bases,
#        num_replicates=2000).sfs
#
#    sfs_len = sfs.n_nonzero_entries
#
#    print("total entries", sfs_len)
#    print("total snps", sfs.n_snps())
#
#    assert sfs_len > 30
#
#    v = np.random.normal(size=len(x0))
#
#    hess1 = hessian_vector_product(SfsLikelihoodSurface(
#        sfs, batch_size=-1, demo_func=demo_func, mut_rate=mu).log_lik)(x0, v)
#    hess2 = hessian_vector_product(lambda x: momi.likelihood._composite_log_likelihood(
#        sfs, demo_func(*x), mut_rate=mu))(x0, v)
#    #hess1 = hessian(SfsLikelihoodSurface(
#    #    sfs, batch_size=-1, demo_func=demo_func, mut_rate=mu).log_lik)(x0)
#    #hess2 = hessian(lambda x: momi.likelihood._composite_log_likelihood(
#    #    sfs, demo_func(*x), mut_rate=mu))(x0)
#    assert np.allclose(hess1, hess2)



import json
import os
import momi
import hashlib
from momi.data.compressed_counts import CompressedAlleleCounts
from io import StringIO

from demo_utils import simple_five_pop_demo
import autograd.numpy as np
from collections import Counter


def test_combine_loci():
    demo = simple_five_pop_demo()
    n_loci = 1000

    num_bases = 1000
    mu = .1
    data = demo.simulate_data(
        num_bases, recoms_per_gen=0,
        num_replicates=n_loci,
        muts_per_gen=mu/num_bases,
        sampled_n_dict=dict(zip(demo.leafs, [10]*5)))

    data.extract_sfs(10).combine_loci()


def test_import_dadi_sfs():
    folded_sfs = "test_dadi_folded_2pop.sfs"
    unfolded_sfs = "test_dadi_unfolded_2pop.sfs"

    folded = momi.data.convert.sfs_from_dadi(folded_sfs, outfile="test_folded.sfs")
    unfolded = momi.data.convert.sfs_from_dadi(unfolded_sfs, outfile="test_unfolded.sfs")

    ## Assert the n_snps in each sfs is the number in the vcf file we created the sfs from.
    assert(folded.n_snps() == 938)
    assert(unfolded.n_snps() == 938)

#def test_load_data():
#    data_path = "test_data.json"
#    if not os.path.exists(data_path):
#        with open(data_path, "w") as data_f:
#            vcf_path = "test_vcf.vcf"
#            data = momi.SnpAlleleCounts.read_vcf(
#                vcf_path, ind2pop={
#                    f"msp_{i}": f"Pop{i%3}"
#                    for i in range(1,7)})
#            data.dump(data_f)
#
#
#    with open(data_path) as f:
#        data = momi.SnpAlleleCounts.load(f)
#
#    with open(data_path) as f:
#        info = json.load(f)
#
#        chrom_pos_config_key = "(chrom_id,position,config_id)"
#        chrom_ids, positions, config_ids = zip(
#            *info[chrom_pos_config_key])
#        del info[chrom_pos_config_key]
#
#        compressed_counts = CompressedAlleleCounts(
#            np.array(info["configs"], dtype=int),
#            np.array(config_ids, dtype=int))
#        del info["configs"]
#
#        #data2 =  momi.SnpAlleleCounts(
#        #    chrom_ids, positions, compressed_counts,
#        #    non_ascertained_pops=[], **info)
#        data2 = momi.snp_allele_counts(
#            chrom_ids, positions, info["populations"],
#            (config[:,0] for config in compressed_counts),
#            (config[:,1] for config in compressed_counts),
#            length=info["length"],
#            use_folded_sfs=info["use_folded_sfs"])
#
#    assert data._sfs == data2._sfs




import numpy as np
import momi.moran_model as moran_model
import pytest
from autograd import grad
from autograd.numpy import dot
import numdifftools as nd


def test_rate_matrix():
    n = 4
    M = moran_model.rate_matrix(n).todense()
    assert np.allclose(M,
                       [[0, 0, 0, 0, 0],
                        [(n - 1) / 2, -(n - 1), (n - 1) / 2, 0, 0],
                           [0, 2 * (n - 2) / 2, -2 * (n - 2),
                            2 * (n - 2) / 2, 0],
                           [0, 0, 3 * (n - 3) / 2, -3 * (n - 3),
                            3 * (n - 3) / 2],
                           [0, 0, 0, 0, 0]])

# @pytest.mark.parametrize("n,t",
#         ((n, t) for n in (5, 10, 50, 100, 250)
#             for t in (0.01, 0.1, 1.0, 10.0, 100.0) if n * t < 100))
# def test_eig_vs_expm(n, t):
#     print(n, t)

#     v = np.random.random(n + 1)

#     x,y,z = 1.0,2.0,3.0

#     t = x**2 * y**3 * z**4 * t

#     v = x**3 * y**2 * z**3 * v

#     w1 = moran_model.moran_action_eigen(t, v)
#     w2 = moran_model.moran_al_mohy_higham(t,v)

#     assert np.allclose(w1, w2)

#     def mma(y):
#         return dot(moran_model.moran_action(t,y), y)
#     mmagrad = grad(mma)
#     mmaegrad = nd.Jacobian(mma)
#     assert np.max(np.log(np.array(mmagrad(v)) / np.array(mmaegrad(v)))) < 1e-8

#     ## hessian vector product
#     mmahess_dot_y = grad(lambda x: dot(mmagrad(x),x))
#     mmahess_dot_y_e = nd.Jacobian(lambda x: dot(mmagrad(x),x))

#     #mmagrad_dot_y(v)
#     assert np.max(np.log(mmahess_dot_y(v) / mmahess_dot_y_e(v))) < 1e-6
